"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PatchPack = void 0;
const notepack = __importStar(require("notepack.io"));
const common_1 = require("./common");
const schema_1 = require("./schema");
class PatchPack {
    constructor(schema) {
        this.schema = new schema_1.Schema(schema);
    }
    encodeSchemaPatch(patch) {
        // split path to elements
        // "/nodes/1/2"
        const [, type, index, prop] = patch.path.split("/");
        // set operation for patch type
        // if "/nodes/" updated op = -3 | -2 | -1
        // if "/types/" updated op = -6 | -5 | -4
        const opIndex = ["add", "replace", "remove"].indexOf(patch.op) - (type === "nodes" ? 3 : 6);
        // return SchemaPatch
        return notepack.encode([opIndex, +index, prop === undefined ? -1 : +prop, patch.value]);
    }
    encodeNode(snapshot, sn, name) {
        const node = name === undefined ? sn : this.schema.getChildNode(sn, name);
        if (!node || !snapshot) {
            return snapshot;
        }
        if (node.type === -2) {
            return node.items.map((key, i) => this.encodeNode(snapshot[key], node, key));
        }
        else if (node.type === -1) {
            return snapshot.map((item, i) => this.encodeNode(item, node, i));
        }
        else {
            const props = this.schema.getTypeProps(node.type);
            return props.map((key, i) => this.encodeNode(snapshot[key], node, key));
        }
    }
    encodeSnapshot(snapshot, nodeId) {
        const node = this.schema.getNode(nodeId);
        if (!node) {
            throw new Error(`Cannot encode Snapshot - node with id ${nodeId} not found!`);
        }
        return notepack.encode(this.encodeNode(snapshot, node));
    }
    encodePatch(patch) {
        const pathArr = patch.path === "/" ? [""] : patch.path.split("/");
        let i = -1;
        let node;
        while (++i < pathArr.length) {
            const child = this.schema.getChildNode(node, pathArr[i] || "");
            if (!child) {
                break;
            }
            node = child;
        }
        if (!node) {
            throw new Error(`Wrong patch path: ${patch.path}`);
        }
        const op = ["add", "replace", "remove"].indexOf(patch.op);
        const props = this.schema.getTypeProps(node.type);
        const propIndex = props.indexOf(pathArr[i]);
        const data = [op, node.id, propIndex];
        if (patch.op !== "remove") {
            data.push(this.encodeNode(patch.value, node, pathArr[i]));
        }
        if (patch.op !== "add" && "oldValue" in patch) {
            data.push(this.encodeNode(patch.oldValue, node, pathArr[i]));
        }
        return notepack.encode(data);
    }
    decodeSchemaPatch(sp, prefix = "") {
        const patch = common_1.schemaPatch(sp);
        // set JsonPatch operation
        const op = ["add", "replace", "remove"][patch.op + (patch.op < -3 ? 6 : 3)];
        // set path elements
        const pathArr = [, patch.op < -3 ? "types" : "nodes", patch.id];
        if (patch.prop >= 0) {
            pathArr.push(patch.prop);
        }
        const jsonPatch = { op, path: prefix + pathArr.join("/") };
        if (patch.values[0] !== undefined) {
            jsonPatch.value = patch.values[0];
        }
        // return JsonPatch
        return jsonPatch;
    }
    decodeNode(nodeId, encoded) {
        const node = this.schema.getNode(nodeId);
        if (!node || !encoded) {
            return encoded;
        }
        if (node.type === common_1.NODE_ARRAY_TYPE) {
            return encoded.map((item, i) => {
                const itemNode = this.schema.getChildNode(node, i);
                return this.decodeNode(itemNode ? itemNode.id : -1, item);
            });
        }
        else {
            const result = {};
            const keys = node.type === common_1.NODE_MAP_TYPE ? node.items : this.schema.getTypeProps(node.type);
            keys.forEach((key, i) => {
                const itemNode = this.schema.getChildNode(node, key);
                const value = this.decodeNode(itemNode ? itemNode.id : -1, encoded[i]);
                if (value === undefined) {
                    return;
                }
                result[key] = value;
            });
            return result;
        }
    }
    decodeSnapshot(buffer, nodeId) {
        return this.decodeNode(nodeId, notepack.decode(buffer));
    }
    decodePatch(buffer, updateSchema = true) {
        const encodedPatch = notepack.decode(buffer);
        if (encodedPatch[0] < 0) {
            // decode schemaMap patch
            const decodedPatch = this.decodeSchemaPatch(encodedPatch, "");
            if (updateSchema) {
                const pathArr = decodedPatch.path.split("/");
                const type = pathArr[1];
                const index = +pathArr[2];
                switch (decodedPatch.op) {
                    case "add":
                        return pathArr.length === 4
                            ? this.schema.addMapNodeKey(+pathArr[2], encodedPatch[3])
                            : this.schema.add(type, encodedPatch[3], index);
                    case "remove":
                        return this.schema.remove(type, index);
                    case "replace":
                        return this.schema.replace(type, index, encodedPatch[3]);
                }
            }
            return decodedPatch;
        }
        const [opIndex, entryId, propIndex, ...values] = encodedPatch;
        const node = this.schema.getNode(entryId);
        if (!node) {
            throw new Error(`Cannot decode patch - schema for node with id ${entryId} not found`);
        }
        const path = this.schema.getNodePath(node);
        const props = this.schema.getTypeProps(node.type);
        const patch = {
            op: ["add", "replace", "remove"][opIndex],
            path: path + (propIndex >= 0 ? "/" + props[propIndex] : "")
        };
        const child = this.schema.findNode({ parent: entryId, index: propIndex });
        const nodeId = child ? child.id : propIndex < 0 && entryId || -1;
        if (values.length && patch.op !== "remove") {
            patch.value = this.decodeNode(nodeId, values.reverse().pop());
        }
        if (values.length && patch.op !== "add") {
            patch.oldValue = this.decodeNode(nodeId, values.pop());
        }
        if (values.length) {
            throw new Error(`Unhandled params: ${values.toString()}`);
        }
        return patch;
    }
}
exports.PatchPack = PatchPack;
